import contextvars

current_engine = contextvars.ContextVar("current_engine")


def get_current_engine():
    try:
        return current_engine.get()
    except LookupError as e:
        raise Exception("No engine set. Did you enter the context?") from e


def set_current_engine(engine):
    current_engine.set(engine)



import dataclasses

from akuire.managers.base import Manager


@dataclasses.dataclass
class SystemConfig:
    """Configuration for the system.

    A system is composed of multiple managers. This class holds the configuration for the system.
    , in the form of a dictionary of managers.



    """

    managers: dict[str, Manager]



import asyncio
import dataclasses
from typing import AsyncGenerator

import numpy as np

from akuire.events import (
    AcquireFrameEvent,
    AcquireZStackEvent,
    DataEvent,
    ImageDataEvent,
    ManagerEvent,
    ZChangeEvent,
)
from akuire.managers.base import Manager


@dataclasses.dataclass
class SweepableManager(Manager):

    async def compute_event(
        self, event: AcquireFrameEvent | ZChangeEvent | AcquireZStackEvent
    ) -> AsyncGenerator[DataEvent, None]:

        if isinstance(event, AcquireFrameEvent):
            print("Acquiring frame")
            yield ImageDataEvent(data=np.random.rand(1, 1, 1, 512, 512, 1))

        if isinstance(event, ZChangeEvent):
            print("Changing z")
            yield ZChangeEvent(z=event.z)

        if isinstance(event, AcquireZStackEvent):
            print("Streaming Z-Stack")
            yield ImageDataEvent(data=np.random.rand(1, 1, event.z_steps, 512, 512))

    def challenge(self, event: ManagerEvent) -> bool:
        if isinstance(event, AcquireFrameEvent):
            return True
        if isinstance(event, ZChangeEvent):
            return True
        if isinstance(event, AcquireZStackEvent):
            return True
        return False



import asyncio
import dataclasses
from typing import AsyncGenerator

import numpy as np

from akuire.events import (
    DataEvent,
    HasMovedEvent,
    ManagerEvent,
    MoveEvent,
    MoveXEvent,
    MoveYEvent,
)
from akuire.managers.base import Manager


@dataclasses.dataclass
class VirtualStageManager(Manager):

    async def compute_event(
        self, event: MoveXEvent | MoveYEvent | MoveEvent
    ) -> AsyncGenerator[DataEvent, None]:
        print("Moving stage")

        yield HasMovedEvent(x=event.x)

    def challenge(self, event: ManagerEvent) -> bool:
        return isinstance(event, MoveEvent)
        return isinstance(event, MoveEvent)



from akuire.acquisition import Acquisition, AcquisitionResult
from akuire.engine import AcquisitionEngine
from akuire.events import AcquireFrameEvent
from akuire.vars import get_current_engine


async def one_shot_snap(
    exposure_time: float = 3, engine: AcquisitionEngine = None
) -> AcquisitionResult:

    engine = engine or get_current_engine()
    return await engine.acquire(Acquisition(events=[AcquireFrameEvent()]))



import asyncio
import dataclasses
import sys
from functools import reduce
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Generator,
    List,
    Protocol,
    Type,
    runtime_checkable,
)

import numpy as np
from koil import unkoil
from koil.composition.base import KoiledModel

from akuire.acquisition import Acquisition, AcquisitionResult
from akuire.compilers.default import compile_events
from akuire.config import SystemConfig
from akuire.events import (
    DataEvent,
    ImageDataEvent,
    ManagerEvent,
    MoveEvent,
    ZChangeEvent,
)
from akuire.vars import set_current_engine

Hook = Callable[[DataEvent], Awaitable[None]]


class AcquisitionEngine(KoiledModel):
    """Acquisition engine that orchestrates the acquisition of data from multiple devices.

    The AcquisitionEngine is the main entry point for orchestrating the acquisition of data from multiple devices.
    It is responsible for compiling the events that need to be executed, and for executing them in the correct order,
    ensuring that the data is collected in the correct order.

    """

    system_config: SystemConfig
    compiler: Callable[[Acquisition, SystemConfig], List[ManagerEvent]] = compile_events
    _lock: asyncio.Lock | None = None

    async def astart(self):
        pass

    def start(self):
        return unkoil(self.astart)

    async def acquire_stream(self, x: Acquisition) -> AsyncGenerator[DataEvent, None]:

        events_queue = self.compiler(x, self.system_config)
        async with self._lock:
            for paired_events in events_queue:
                for paired_event in paired_events:
                    async for event in self.system_config.managers[
                        paired_event.manager
                    ].compute_event(paired_event.event):
                        yield event

    async def acquire(
        self, x: Acquisition, hooks: list[Hook] | None = None
    ) -> AcquisitionResult:

        collected_events = []
        hooks = hooks or []

        async for event in self.acquire_stream(x):
            collected_events.append(event)
            for hook in hooks:
                await hook(event)

        return AcquisitionResult(collected_events)

    async def __aenter__(self) -> "AcquisitionEngine":
        self._lock = asyncio.Lock()
        set_current_engine(self)

        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        self._lock = None
        set_current_engine(None)
        pass

    class Config:
        arbitrary_types_allowed = True
        underscore_attrs_are_private = True



import dataclasses

import numpy as np


@dataclasses.dataclass
class DataEvent:
    """Base class for all data events

    Data events are used to communicate data that are
    generated during the acquisition process. This
    can be images, metadata, or other data that is
    generated during the acquisition process.


    """

    pass


@dataclasses.dataclass
class ImageDataEvent(DataEvent):

    data: np.ndarray
    """Data should be 5 Dimensional c,t,z,y,x"""


@dataclasses.dataclass
class HasMovedEvent(DataEvent):
    x: float | None = None
    y: float | None = None
    z: float | None = None
    pass
    pass



import dataclasses
from typing import List

import numpy as np

from akuire.events import ManagerEvent
from akuire.events.data_event import DataEvent, ImageDataEvent


@dataclasses.dataclass
class PairedEvent:
    manager: str
    event: ManagerEvent


@dataclasses.dataclass
class Acquisition:
    """Class that represents an acquisition

    An acquisition is a collection of events that are executed in sequence.
    The events can be of different types and are executed in the order they are provided.

    While the events are executed in sequence, some events might be transpiled
    into multiple events if the underlying manager does not support the event.

    E.g if the manager does not support a Z-stack acquisition (i.e it cant optimize
    this into a single event that acquires a Z-stack in one go), the AcquireZStackEvent
    will be transpiled into multiple MoveZEvents and AcquireFramevents.

    """

    events: List[ManagerEvent]
    context: dict = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class AcquisitionResult:
    """Class to store the result of an acquisition

    This class is used to store the result of an acquisition. It contains the collected events and
    provides a method to stack the collected ImageDataEvents into a single Z-stack.

    Attributes:
        collected_events (List[DataEvent]): List of DataEvents that were collected during the acquisition





    """

    collected_events: List[DataEvent]

    def to_z_stack(self):
        """Stack the collected ImageDataEvents into a single Z-stack

        This method stacks the collected ImageDataEvents into a single Z-stack. It will raise a ValueError if none of the
        collected events are ImageDataEvents.
        """
        try:
            return np.stack(
                [
                    x.data
                    for x in self.collected_events
                    if isinstance(x, ImageDataEvent)
                ],
                axis=0,
            )
        except ValueError as e:
            print(self.collected_events)
            raise ValueError(
                "Not all collected events are ImageDataEvents, cannot stack. Check the collected events."
            ) from e



import asyncio
import dataclasses
from typing import AsyncGenerator

import cv2
import matplotlib.pyplot as plt
import NanoImagingPack as nip
import numpy as np
from scipy.signal import convolve2d

# Use the line function from skimage
from skimage.draw import line

from akuire.assets.utils import get_absolute_path
from akuire.events import (
    AcquireFrameEvent,
    AcquireZStackEvent,
    DataEvent,
    HasMovedEvent,
    ImageDataEvent,
    ManagerEvent,
    MoveEvent,
    SetLightIntensityEvent,
    ZChangeEvent,
)
from akuire.events.manager_event import AcquireTSeriesEvent
from akuire.managers.base import Manager

try:
    import NanoImagingPack as nip

    IS_NIP = True
except:
    IS_NIP = False
import math
import os
import threading
import time

import cv2
import matplotlib.pyplot as plt
import numpy as np

# workaround to not break if numba is not installed
try:
    from numba import njit, prange
except ModuleNotFoundError:
    prange = range

    def njit(*args, **kwargs):
        def wrapper(func):
            return func

        return wrapper


# workaroung to not break if nanopyx is not installed
try:
    from nanopyx import eSRRF
except ModuleNotFoundError:

    def eSRRF(img, **kwargs):
        # if nanopyx is not installed just returns the input image
        return img


IS_NIP = True


FILE_PATH = get_absolute_path("big.jpg")


@njit(parallel=True)
def FromLoc2Image_MultiThreaded(
    xc_array, yc_array, photon_array, sigma_array, image_size, pixel_size
):
    """
    Function to generate an image from a list of emitter locations.
    Based on the DeepStorm notebook on ZeroCostDL4Mic.
    https://colab.research.google.com/github/HenriquesLab/ZeroCostDL4Mic/blob/master/Colab_notebooks/Deep-STORM_2D_ZeroCostDL4Mic.ipynb
    """
    Image = np.zeros((image_size, image_size))
    for ij in prange(image_size * image_size):
        j = int(ij / image_size)
        i = ij - j * image_size
        for xc, yc, photon, sigma in zip(xc_array, yc_array, photon_array, sigma_array):
            # Don't bother if the emitter has photons <= 0 or if Sigma <= 0
            if (photon > 0) and (sigma > 0):
                S = sigma * math.sqrt(2)
                x = i * pixel_size - xc
                y = j * pixel_size - yc
                # Don't bother if the emitter is further than 4 sigma from the centre of the pixel
                if (x + pixel_size / 2) ** 2 + (
                    y + pixel_size / 2
                ) ** 2 < 16 * sigma**2:
                    ErfX = math.erf((x + pixel_size) / S) - math.erf(x / S)
                    ErfY = math.erf((y + pixel_size) / S) - math.erf(y / S)
                    Image[j][i] += 0.25 * photon * ErfX * ErfY
    return Image


@dataclasses.dataclass
class Position:
    x: float = 0
    y: float = 0
    z: float = 0


@dataclasses.dataclass
class SMLMMicroscope(Manager):
    """A virtual SMLM microscope.

    This class simulates a virtual SMLM microscope. It generates images based on the current settings.


    """

    sensor_height: int = 512
    """The height of the sensor in pixels."""
    sensor_width: int = 512
    """The width of the sensor in pixels."""
    NA: float = 1.2
    """The numerical aperture of the microscope."""
    pixel_size: float = 0.1
    magnification: float = 2
    readout_noise: float = 50

    density: float = 0.05
    ADC_per_photon_conversion: float = 1.0
    ADC_offset: float = 100

    wave_length: float = 0.6
    wave_length_std: float = 0.5

    position: Position = dataclasses.field(default_factory=Position)
    intensity: float = 0.1
    intensity_std_dev: float = 0.01
    dz: float = 0
    image: np.ndarray = dataclasses.field(
        default_factory=lambda: createBranchingTree(width=512, height=512)
    )

    def produce_smlm_frame(
        self,
        x_offset: int = 0,
        y_offset: int = 0,
        n_photons: int = 5000,
        n_photons_std: int = 50,
    ):
        """Generate a frame based on the current settings."""
        # add moise
        image = self.image.copy()
        # Adjust image based on offsets
        image = np.roll(np.roll(image, int(x_offset), axis=1), int(y_offset), axis=0)
        image = nip.extract(image, (self.sensor_width, self.sensor_height))

        yc_array, xc_array = self.binary2locs(image, density=self.density)
        photon_array = np.random.normal(n_photons, n_photons_std, size=len(xc_array))

        wavelenght = self.wave_length
        wavelenght_std = self.wave_length_std
        NA = self.NA
        sigma = 0.21 * wavelenght / NA  # change to get it from microscope settings
        sigma_std = (
            0.21 * wavelenght_std / NA
        )  # change to get it from microscope settings
        sigma_array = np.random.normal(sigma, sigma_std, size=len(xc_array))

        ADC_per_photon_conversion = (
            self.ADC_per_photon_conversion
        )  # change to get it from microscope settings
        readout_noise = self.readout_noise  # change to get it from microscope settings
        ADC_offset = self.ADC_offset  # change to get it from microscope settings

        out = FromLoc2Image_MultiThreaded(
            xc_array,
            yc_array,
            photon_array,
            sigma_array,
            self.sensor_height,
            self.pixel_size,
        )
        out = (
            ADC_per_photon_conversion * np.random.poisson(out)
            + readout_noise
            * np.random.normal(size=(self.sensor_height, self.sensor_width))
            + ADC_offset
        )
        return np.array(out)

    def binary2locs(self, img: np.ndarray, density: float):
        """
        Given a binary image `img` and a `density` value, this function returns a subset of the locations where the image is non-zero.
        Parameters:
            img (np.ndarray): A binary image.
            density (float): The density of the subset of locations to be returned.
        Returns:
            tuple: A tuple containing two numpy arrays representing the x and y coordinates of the subset of locations.
        """
        all_locs = np.nonzero(img == 1)
        n_points = int(len(all_locs[0]) * density)
        selected_idx = np.random.choice(len(all_locs[0]), n_points, replace=False)
        filtered_locs = all_locs[0][selected_idx], all_locs[1][selected_idx]
        return filtered_locs

    async def get_last(self) -> np.ndarray:

        return await asyncio.to_thread(
            self.produce_smlm_frame,
            x_offset=self.position.x,
            y_offset=self.position.y,
            n_photons=self.intensity,
            n_photons_std=self.intensity * self.intensity_std_dev,
        )

    async def compute_event(
        self, event: AcquireFrameEvent | ZChangeEvent
    ) -> AsyncGenerator[DataEvent, None]:

        if isinstance(event, AcquireFrameEvent):
            print("Acquiring frame")
            yield ImageDataEvent(data=await self.get_last())

        if isinstance(event, SetLightIntensityEvent):
            print("Setting light intensity")
            self.intensity = event.intensity

        if isinstance(event, AcquireTSeriesEvent):
            print("Acquiring Z-stack")
            for z in range(event.t_steps):
                yield ImageDataEvent(data=await self.get_last())

        if isinstance(event, AcquireZStackEvent):
            print("Acquiring Z-stack")
            z_stack = []
            for z in event.z_steps:
                self.position.z = z
                yield ImageDataEvent(data=await self.get_last())

        if isinstance(event, MoveEvent):
            print("Moving")
            if event.x:
                self.position.x = event.x
            if event.y:
                self.position.y = event.y
            if event.z:
                self.position.z = event.z

            yield HasMovedEvent(x=self.position.x, y=self.position.y, z=self.position.z)

    def challenge(self, event: ManagerEvent) -> bool:
        return isinstance(
            event,
            (
                AcquireFrameEvent,
                SetLightIntensityEvent,
                AcquireZStackEvent,
                MoveEvent,
                AcquireTSeriesEvent,
            ),
        )


@dataclasses.dataclass
class Camera:
    sensor_height: int = 512
    sensor_width: int = 512
    pixel_size: float = 1.0
    image: np.ndarray = dataclasses.field(
        default_factory=lambda: createBranchingTree(width=512, height=512)
    )

    def __post_init__(self):
        self.noise_stack = (
            np.random.randn(self.sensor_height, self.sensor_width, 100) * 2
        )

    def setPropertyValue(self, propertyName, propertyValue):
        pass


def createBranchingTree(width=5000, height=5000, lineWidth=3):
    np.random.seed(0)  # Set a random seed for reproducibility
    # Define the dimensions of the image
    width, height = 5000, 5000

    # Create a blank white image
    image = np.ones((height, width), dtype=np.uint8) * 255

    # Function to draw a line (blood vessel) on the image
    def draw_vessel(start, end, image):
        rr, cc = line(start[0], start[1], end[0], end[1])
        try:
            image[rr, cc] = 0  # Draw a black line
        except:
            end = 0
            return

    # Recursive function to draw a tree-like structure
    def draw_tree(start, angle, length, depth, image, reducer, max_angle=40):
        if depth == 0:
            return

        # Calculate the end point of the branch
        end = (
            int(start[0] + length * np.sin(np.radians(angle))),
            int(start[1] + length * np.cos(np.radians(angle))),
        )

        # Draw the branch
        draw_vessel(start, end, image)

        # change the angle slightly to add some randomness
        angle += np.random.uniform(-10, 10)

        # Recursively draw the next level of branches
        new_length = length * reducer  # Reduce the length for the next level
        new_depth = depth - 1
        draw_tree(
            end,
            angle - max_angle * np.random.uniform(-1, 1),
            new_length,
            new_depth,
            image,
            reducer,
        )
        draw_tree(
            end,
            angle + max_angle * np.random.uniform(-1, 1),
            new_length,
            new_depth,
            image,
            reducer,
        )

    # Starting point and parameters
    start_point = (height - 1, width // 2)
    initial_angle = -90  # Start by pointing upwards
    initial_length = np.max((width, height)) * 0.15  # Length of the first branch
    depth = 7  # Number of branching levels
    reducer = 0.9
    # Draw the tree structure
    draw_tree(start_point, initial_angle, initial_length, depth, image, reducer)

    # convolve image with rectangle
    rectangle = np.ones((lineWidth, lineWidth))
    image = convolve2d(image, rectangle, mode="same", boundary="fill", fillvalue=0)

    return image



import asyncio
import dataclasses
from typing import AsyncGenerator

import cv2
import matplotlib.pyplot as plt
import NanoImagingPack as nip
import numpy as np
from scipy.signal import convolve2d

# Use the line function from skimage
from skimage.draw import line

from akuire.assets.utils import get_absolute_path
from akuire.events import (
    AcquireFrameEvent,
    AcquireTSeriesEvent,
    AcquireZStackEvent,
    DataEvent,
    ImageDataEvent,
    ManagerEvent,
    MoveEvent,
    SetLightIntensityEvent,
    ZChangeEvent,
)
from akuire.managers.base import Manager

IS_NIP = True


FILE_PATH = get_absolute_path("big.jpg")


class VirtualMicroscopeManager(Manager):

    def __init__(self, filePath=FILE_PATH):
        self.camera = Camera(self, filePath)
        self.positioner = Positioner(self)
        self.illuminator = Illuminator(self)

    async def compute_event(
        self, event: AcquireFrameEvent | ZChangeEvent
    ) -> AsyncGenerator[DataEvent, None]:

        if isinstance(event, AcquireFrameEvent):
            print("Acquiring frame")
            yield ImageDataEvent(data=self.camera.getLast())

        if isinstance(event, SetLightIntensityEvent):
            print("Setting light intensity")
            self.illuminator.set_intensity(event.intensity)

        if isinstance(event, AcquireZStackEvent):
            print("Acquiring Z-stack")
            z_stack = []
            for z in event.z_values:
                self.positioner.move(z=z)
                z_stack.append(self.camera.getLast())
            yield ImageDataEvent(data=z_stack)

        if isinstance(event, MoveEvent):
            print("Moving")
            self.positioner.move(x=event.x, y=event.y, z=event.z)

    def challenge(self, event: ManagerEvent) -> bool:
        return isinstance(
            event,
            (
                AcquireFrameEvent,
                SetLightIntensityEvent,
                AcquireZStackEvent,
                MoveEvent,
                AcquireTSeriesEvent,
            ),
        )


class Camera:
    def __init__(self, parent, filePath="simplant"):
        self._parent = parent
        if filePath == "simplant":
            self.image = createBranchingTree(width=5000, height=5000)
        else:
            self.image = np.mean(cv2.imread(filePath), axis=2)

        self.image /= np.max(self.image)
        self.SensorWidth = 512  # self.image.shape[1]
        self.SensorHeight = 512  # self.image.shape[0]
        self.model = "VirtualCamera"
        self.PixelSize = 1.0
        self.isRGB = False
        self.frameNumber = 0
        # precompute noise so that we will save energy and trees
        self.noiseStack = np.random.randn(self.SensorHeight, self.SensorWidth, 100) * 2

    def produce_frame(
        self, x_offset=0, y_offset=0, light_intensity=1.0, defocusPSF=None
    ):
        """Generate a frame based on the current settings."""
        # add moise
        image = self.image.copy()
        # Adjust image based on offsets
        image = np.roll(np.roll(image, int(x_offset), axis=1), int(y_offset), axis=0)
        image = nip.extract(image, (self.SensorWidth, self.SensorHeight))

        # do all post-processing on cropped image
        if IS_NIP and defocusPSF is not None and not defocusPSF.shape == ():
            print("Defocus:" + str(defocusPSF.shape))
            image = np.array(np.real(nip.convolve(image, defocusPSF)))
        image = np.float32(image) / np.max(image) * np.float32(light_intensity)
        image += self.noiseStack[:, :, np.random.randint(0, 100)]

        # Adjust illumination
        image = image.astype(np.uint16)
        print(image.shape)
        reshaped = np.array(image).reshape(
            (1, 1, 1, self.SensorHeight, self.SensorWidth)
        )
        print(reshaped.shape)
        return reshaped

    def getLast(self, returnFrameNumber=False):
        position = self._parent.positioner.get_position()
        defocusPSF = np.squeeze(self._parent.positioner.get_psf())
        intensity = self._parent.illuminator.get_intensity(1)
        self.frameNumber += 1
        if returnFrameNumber:
            return (
                self.produce_frame(
                    x_offset=position["X"],
                    y_offset=position["Y"],
                    light_intensity=intensity,
                    defocusPSF=defocusPSF,
                ),
                self.frameNumber,
            )
        else:
            return self.produce_frame(
                x_offset=position["X"],
                y_offset=position["Y"],
                light_intensity=intensity,
                defocusPSF=defocusPSF,
            )

    def setPropertyValue(self, propertyName, propertyValue):
        pass


class Positioner:
    def __init__(self, parent):
        self._parent = parent
        self.position = {"X": 0, "Y": 0, "Z": 0, "A": 0}
        self.mDimensions = (
            self._parent.camera.SensorWidth,
            self._parent.camera.SensorHeight,
        )
        if IS_NIP:
            self.psf = self.compute_psf(dz=0)
        else:
            self.psf = None

    def move(self, x=None, y=None, z=None, a=None, is_absolute=True):
        if is_absolute:
            if x is not None:
                self.position["X"] = x
            if y is not None:
                self.position["Y"] = y
            if z is not None:
                self.position["Z"] = z
                self.compute_psf(self.position["Z"])
            if a is not None:
                self.position["A"] = a
        else:
            if x is not None:
                self.position["X"] += x
            if y is not None:
                self.position["Y"] += y
            if z is not None:
                self.position["Z"] += z
                self.compute_psf(self.position["Z"])
            if a is not None:
                self.position["A"] += a

    def get_position(self):
        return self.position.copy()

    def compute_psf(self, dz):
        dz = np.float32(dz)
        print("Defocus:" + str(dz))
        if IS_NIP and dz != 0:
            obj = nip.image(np.zeros(self.mDimensions))
            obj.pixelsize = (100.0, 100.0)
            paraAbber = nip.PSF_PARAMS()
            # aber_map = nip.xx(obj.shape[-2:]).normalize(1)
            paraAbber.aberration_types = [paraAbber.aberration_zernikes.spheric]
            paraAbber.aberration_strength = [np.float32(dz) / 10]
            psf = nip.psf(obj, paraAbber)
            self.psf = psf.copy()
            del psf
            del obj
        else:
            self.psf = None

    def get_psf(self):
        return self.psf


class Illuminator:
    def __init__(self, parent):
        self._parent = parent
        self.intensity = 100

    def set_intensity(self, intensity=0):
        self.intensity = intensity

    def get_intensity(self, channel):
        return self.intensity


def createBranchingTree(width=5000, height=5000, lineWidth=3):
    np.random.seed(0)  # Set a random seed for reproducibility
    # Define the dimensions of the image
    width, height = 5000, 5000

    # Create a blank white image
    image = np.ones((height, width), dtype=np.uint8) * 255

    # Function to draw a line (blood vessel) on the image
    def draw_vessel(start, end, image):
        rr, cc = line(start[0], start[1], end[0], end[1])
        try:
            image[rr, cc] = 0  # Draw a black line
        except:
            end = 0
            return

    # Recursive function to draw a tree-like structure
    def draw_tree(start, angle, length, depth, image, reducer, max_angle=40):
        if depth == 0:
            return

        # Calculate the end point of the branch
        end = (
            int(start[0] + length * np.sin(np.radians(angle))),
            int(start[1] + length * np.cos(np.radians(angle))),
        )

        # Draw the branch
        draw_vessel(start, end, image)

        # change the angle slightly to add some randomness
        angle += np.random.uniform(-10, 10)

        # Recursively draw the next level of branches
        new_length = length * reducer  # Reduce the length for the next level
        new_depth = depth - 1
        draw_tree(
            end,
            angle - max_angle * np.random.uniform(-1, 1),
            new_length,
            new_depth,
            image,
            reducer,
        )
        draw_tree(
            end,
            angle + max_angle * np.random.uniform(-1, 1),
            new_length,
            new_depth,
            image,
            reducer,
        )

    # Starting point and parameters
    start_point = (height - 1, width // 2)
    initial_angle = -90  # Start by pointing upwards
    initial_length = np.max((width, height)) * 0.15  # Length of the first branch
    depth = 7  # Number of branching levels
    reducer = 0.9
    # Draw the tree structure
    draw_tree(start_point, initial_angle, initial_length, depth, image, reducer)

    # convolve image with rectangle
    rectangle = np.ones((lineWidth, lineWidth))
    image = convolve2d(image, rectangle, mode="same", boundary="fill", fillvalue=0)

    return image



import asyncio
import dataclasses
from typing import AsyncGenerator

import numpy as np

from akuire.events import (
    AcquireFrameEvent,
    AcquireZStackEvent,
    DataEvent,
    ImageDataEvent,
    ManagerEvent,
    ZChangeEvent,
)
from akuire.managers.base import Manager


@dataclasses.dataclass
class ZStageManager(Manager):

    async def compute_event(
        self, event: ZChangeEvent
    ) -> AsyncGenerator[DataEvent, None]:
        print("Changed Z")
        yield ZChangeEvent(z=event.z)

    def challenge(self, event: ManagerEvent) -> bool:
        return isinstance(event, ZChangeEvent)



import dataclasses
from functools import reduce
from operator import add

from akuire.errors import AtomicException


@dataclasses.dataclass
class ManagerEvent:
    pass

    def transpile(self) -> list["ManagerEvent"]:
        raise AtomicException(
            f"{self.__class__.__name__} even is atomic and cannot be transpiled."
        )


@dataclasses.dataclass
class DeviceChangeEvent(ManagerEvent):
    device: str
    state: bool


@dataclasses.dataclass
class AcquireFrameEvent(ManagerEvent):
    exposure_time: float = 100


@dataclasses.dataclass
class ZChangeEvent(ManagerEvent):
    z: float


@dataclasses.dataclass
class AcquireZStackEvent(ManagerEvent):
    z_steps: float
    item_exposure_time: float

    def transpile(self) -> list[ManagerEvent]:
        print("Transpiling z stack")
        x = reduce(
            lambda x, y: list(x) + list(y),
            list(
                zip(
                    [
                        AcquireFrameEvent(exposure_time=self.item_exposure_time)
                        for x in range(self.z_steps)
                    ],
                    [ZChangeEvent(z=i) for i in range(self.z_steps)],
                )
            ),
            [],
        )

        print(x)

        return x


@dataclasses.dataclass
class AcquireTSeriesEvent(ManagerEvent):
    t_steps: float
    interval: float | None = None
    item_exposure_time: float = 100

    def transpile(self) -> list[ManagerEvent]:
        print("Transpiling z stack")

        return [
            AcquireFrameEvent(exposure_time=self.item_exposure_time)
            for t in range(self.t_steps)
        ]


@dataclasses.dataclass
class MoveEvent(ManagerEvent):
    x: float | None = None
    y: float | None = None
    z: float | None = None
    speed: float = 10000

    def transpile(self) -> list[ManagerEvent]:
        return [MoveXEvent(x=self.x), MoveYEvent(y=self.y)]


@dataclasses.dataclass
class MoveXEvent(ManagerEvent):
    x: float


@dataclasses.dataclass
class GetXEvent(ManagerEvent):
    pass


@dataclasses.dataclass
class SetLightIntensityEvent(ManagerEvent):
    intensity: float


@dataclasses.dataclass
class MoveYEvent(ManagerEvent):
    y: float
    pass



import asyncio
import dataclasses
from typing import AsyncGenerator

import numpy as np

from akuire.events import (
    AcquireFrameEvent,
    AcquireZStackEvent,
    DataEvent,
    ImageDataEvent,
    ManagerEvent,
    ZChangeEvent,
)
from akuire.managers.base import Manager


@dataclasses.dataclass
class NonSweepableCamera(Manager):
    exposition_time_is_sleep: bool = False

    async def compute_event(
        self, event: AcquireFrameEvent | ZChangeEvent
    ) -> AsyncGenerator[DataEvent, None]:

        if isinstance(event, AcquireFrameEvent):
            print("Acquiring frame")
            if self.exposition_time_is_sleep:
                await asyncio.sleep(event.exposition_time)
            yield ImageDataEvent(data=np.random.rand(1, 512, 512, 1))

        yield ImageDataEvent(data=np.random.rand(1, 512, 512, 1))

    def challenge(self, event: ManagerEvent) -> bool:
        return isinstance(event, AcquireFrameEvent)



from .data_event import DataEvent, HasMovedEvent, ImageDataEvent
from .manager_event import (
    AcquireFrameEvent,
    AcquireTSeriesEvent,
    AcquireZStackEvent,
    DeviceChangeEvent,
    ManagerEvent,
    MoveEvent,
    MoveXEvent,
    MoveYEvent,
    SetLightIntensityEvent,
    ZChangeEvent,
)

__all__ = [
    "DataEvent",
    "ImageDataEvent",
    "HasMovedEvent",
    "ManagerEvent",
    "DeviceChangeEvent",
    "MoveEvent",
    "MoveXEvent",
    "MoveYEvent",
    "SetLightIntensityEvent",
    "ZChangeEvent",
    "AcquireZStackEvent",
    "AcquireFrameEvent",
    "AcquireTSeriesEvent",
]



from functools import reduce

from akuire.acquisition import Acquisition, PairedEvent
from akuire.config import SystemConfig
from akuire.errors import AtomicException
from akuire.events import ManagerEvent


def recurse_transpile_until_accepted(
    current_event: ManagerEvent, config: SystemConfig
) -> list[PairedEvent]:

    for key, manager in config.managers.items():
        if manager.challenge(current_event):
            return [PairedEvent(manager=key, event=current_event)]

    print(current_event)
    return reduce(
        lambda x, y: x + y,
        [
            recurse_transpile_until_accepted(event, config)
            for event in current_event.transpile()
        ],
        [],
    )


def compile_events(x: Acquisition, config: SystemConfig) -> list[list[PairedEvent]]:

    manager_queue = []

    for event in x.events:
        try:
            parsable_events = recurse_transpile_until_accepted(event, config)
            print(parsable_events)

            manager_queue.append(parsable_events)
        except AtomicException as e:
            raise AtomicException(
                f"Event {event} cannot be transpiled to an event that can be handled by any Manager."
            ) from e

    return manager_queue



from typing import AsyncContextManager, AsyncGenerator, Protocol, runtime_checkable

from akuire.events.data_event import DataEvent
from akuire.events.manager_event import ManagerEvent


@runtime_checkable
class Manager(Protocol):
    device: str = "default"

    async def compute_event(
        self, event: ManagerEvent
    ) -> AsyncGenerator[DataEvent, None]: ...

    def challenge(self, event: ManagerEvent) -> bool:
        """Check if the manager can handle the event

        The challenge method is used to check if the manager can handle a specific event. This method should return True
        if the manager can handle the event and False otherwise.

        The AcquistionEngine will use this method to determine if the manager can handle the event. If the manager can
        handle the event, the compute_event method will be called to execute the event.


        """
        ...



import asyncio
import dataclasses
import io
from typing import AsyncGenerator
from urllib.parse import urlencode

import aiohttp
import numpy as np
import PIL.Image
import urllib3

from akuire.events import (
    AcquireFrameEvent,
    AcquireZStackEvent,
    DataEvent,
    HasMovedEvent,
    ImageDataEvent,
    ManagerEvent,
    MoveEvent,
    MoveXEvent,
    MoveYEvent,
    ZChangeEvent,
)
from akuire.events.manager_event import SetLightIntensityEvent
from akuire.managers.base import Manager


@dataclasses.dataclass
class OpenUC2RestManager(Manager):
    """Manager for the OpenUC2 REST API

    This manager is used to control the OpenUC2 REST API.
    It can be used to control the LED matrix, the positioner and the camera.

    """

    exposition_time_is_sleep: bool = False
    endpoint: str = "https://192.168.137.1:8002/"
    png_snap_endpoint: str = "RecordingController/snapNumpyToFastAPI/"
    set_intensity_endpoint: str = "LEDMatrixController/setIntensity/"
    set_positioner_endpoint: str = "PositionerController/movePositioner"
    ssl: bool = False
    stage = "ESP32Stage"

    async def compute_event(
        self, event: AcquireFrameEvent | ZChangeEvent | MoveXEvent | MoveYEvent
    ) -> AsyncGenerator[DataEvent, None]:

        if isinstance(event, AcquireFrameEvent):
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    self.endpoint + self.png_snap_endpoint,
                    ssl=self.ssl,
                ) as response:
                    # Download the file
                    # Convert the response to a PIL image
                    bytes_array = await response.read()

                    # Bytes array to image
                    image = PIL.Image.open(io.BytesIO(bytes_array))
                    # to numpy array
                    image = np.array(image)
                    x = np.reshape(image, (1, 1, 1, *image.shape))
                    print(x)
                    yield ImageDataEvent(data=x)

        if isinstance(event, SetLightIntensityEvent):
            async with aiohttp.ClientSession() as session:
                async with session.get(
                    self.endpoint
                    + self.set_intensity_endpoint
                    + "?intensity="
                    + str(int(event.intensity * 255)),
                    ssl=self.ssl,
                ) as response:
                    # Download the file
                    t = await response.read()
                    print(t)

        if isinstance(event, MoveEvent):
            async with aiohttp.ClientSession() as session:

                if event.x is not None:
                    querystring = urlencode(
                        [
                            ("positionerName", self.stage),
                            ("axis", "X"),
                            ("dist", event.x),
                            ("speed", event.speed),
                            ("isBlocking", True),
                            ("isAbsolute", True),
                        ]
                    )

                    url = (
                        self.endpoint + self.set_positioner_endpoint + "?" + querystring
                    )

                    print(url)

                    async with session.get(
                        url,
                        ssl=self.ssl,
                    ) as response:
                        # Download the file
                        t = await response.read()
                        print(t)

                if event.y is not None:
                    querystring = urlencode(
                        [
                            ("positionerName", self.stage),
                            ("axis", "Y"),
                            ("dist", event.y),
                            ("speed", event.speed),
                            ("isBlocking", True),
                            ("isAbsolute", True),
                        ]
                    )

                    url = (
                        self.endpoint + self.set_positioner_endpoint + "?" + querystring
                    )

                    print(url)

                    async with session.get(
                        url,
                        ssl=self.ssl,
                    ) as response:
                        # Download the file
                        t = await response.read()
                        print(t)

                yield HasMovedEvent(x=event.x, y=event.y)

    def challenge(self, event: ManagerEvent) -> bool:
        if isinstance(event, AcquireFrameEvent):
            return True
        if isinstance(event, ZChangeEvent):
            return True
        if isinstance(event, MoveXEvent):

            return True

        if isinstance(event, MoveEvent):
            return True
        if isinstance(event, MoveYEvent):
            return True
        if isinstance(event, SetLightIntensityEvent):
            return True
